"""

    Copyright (c) 2017-2021 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import argparse
import logging
import os
import sys

import dirtreedigest.__config__ as dtconfig
import dirtreedigest.digester as dtdigester
import dirtreedigest.utils as dtutils
import dirtreedigest.walker as dtwalker


def validate_args(headline):
    """ Validate command-line arguments """
    control_data = dtconfig.CONTROL_DATA
    # package_data = dtconfig.PACKAGE_DATA

    digest_list = sorted(dtdigester.DIGEST_FUNCTIONS.keys())
    epilog = "Digests available: {}".format(digest_list)

    parser = argparse.ArgumentParser(
        # description=package_data['description'],
        epilog=epilog)
    parser.add_argument('root', nargs='?', metavar='ROOTPATH',
                        default=None, type=str, action='store',
                        help='root directory for processing')
    parser.add_argument('--digests', dest='selected_digests',
                        metavar='DIGEST1[,DIGEST2...]',
                        default=','.join(control_data['default_digests']),
                        type=str, action='store',
                        help='digests to use')
    parser.add_argument('--altdigest', dest='altfile_digest', metavar='DIGEST',
                        default=None, type=str, action='store',
                        help='alternate single digest report')
    parser.add_argument('--title', dest='output_title', metavar='TITLE',
                        default=None, type=str, action='store',
                        help='alternate output title')
    parser.add_argument('--tstamp', dest='output_tstamp', metavar='TIMESTAMP',
                        default=None, type=str, action='store',
                        help='alternate output timestamp')
    parser.add_argument('--blocksize', dest='blocksize', metavar='MBYTES',
                        default=control_data['max_block_size_mb'], type=int, action='store',
                        help='block size in MB')
    parser.add_argument('--buffers', dest='buffers', metavar='N',
                        default=control_data['max_buffers'], type=int, action='store',
                        help='number of buffers')
    parser.add_argument('--noshm', dest='noshm',
                        action='store_true',
                        help='don\'t use shared memory')
    parser.add_argument('--nocase', dest='nocase',
                        action='store_true',
                        help='case insensitive matching')
    parser.add_argument('--debug', dest='debug',
                        action='store_true',
                        help='more debugging to the logfile')
    parser.add_argument('--xfiles', dest='excluded_files', metavar='FILE1[,FILE2...]',
                        default=None, type=str, action='append',
                        help='excluded files')
    parser.add_argument('--xdirs', dest='excluded_dirs', metavar='DIR1[,DIR2...]',
                        default=None, type=str, action='append',
                        help='excluded directories')
    args = parser.parse_args()

    if not args.root:
        parser.print_help()
        return False

    control_data['logfile_level'] = logging.INFO
    if args.debug:
        control_data['logfile_level'] = logging.DEBUG

    control_data['root_dir'] = dtutils.unixify_path(os.path.realpath(args.root))

    control_data['outfile_suffix'] = control_data['root_dir'].replace(':', '$').replace('/', '_')

    if args.output_title:
        output_title = args.output_title
    else:
        output_title = '{}-{}'.format(
            control_data['outfile_prefix'],
            control_data['outfile_suffix'],
        )

    if args.output_tstamp:
        output_tstamp = args.output_tstamp
    else:
        output_tstamp = dtutils.datetime_as_str()

    control_data['outfile_name'] = '{}.{}.{}'.format(
        output_title,
        output_tstamp,
        control_data['outfile_ext'],
    )

    control_data['logfile_name'] = '{}.{}.{}'.format(
        output_title,
        output_tstamp,
        control_data['logfile_ext'],
    )

    dtutils.start_logging(
        control_data['logfile_name'],
        control_data['logfile_level'],
        control_data['console_level'],
    )

    logger = logging.getLogger('_main_')
    logger.info('Log begins')
    logger.info('-' * 78)
    logger.info(headline)
    logger.info(f"Using Python {sys.version}")
    logger.info('-' * 78)

    logger.info('Root dir (gvn): %s', args.root)
    logger.info('Root dir (mod): %s', control_data['root_dir'])
    if not os.path.isdir(os.path.realpath(args.root)):
        logger.error('Root dir is not a directory / does not exist!')
        return False

    if not 1 <= args.blocksize < 1024:
        logger.error('Block size must be >= 1MB and < 1024 MB')
        return False
    control_data['max_block_size_mb'] = args.blocksize
    control_data['max_block_size'] = args.blocksize * 1024 * 1024
    logger.info('max_block_size: %d MB', control_data['max_block_size_mb'])

    if not 2 <= args.buffers <= 32:
        logger.error('Number of buffers must be >= 2 and <= 32')
        return False
    control_data['max_buffers'] = args.buffers
    logger.info('max_buffers: %d', control_data['max_buffers'])

    control_data['shm_mode'] = True
    if args.noshm or not dtutils.shared_memory_available():
        control_data['shm_mode'] = False
    logger.info('shm_mode: %s', control_data['shm_mode'])

    control_data['ignore_path_case'] = False
    if args.nocase:
        control_data['ignore_path_case'] = True
    logger.info('ignore_path_case: %s', control_data['ignore_path_case'])

    if args.selected_digests:
        arg_mod = ' '.join(args.selected_digests.replace(',', ' ').split())
        control_data['selected_digests'] = arg_mod.split(' ')
    control_data['selected_digests'] = dtdigester.validate_digests(control_data=control_data)
    if not control_data['selected_digests']:
        logger.error('No valid digests selected')
        return False
    logger.info('digests to run: %s', ', '.join(control_data['selected_digests']))

    if args.excluded_files:
        for val in args.excluded_files:
            control_data['ignored_files'].append(val)
    logger.info('ignored_files: %s', ', '.join(control_data['ignored_files']))

    if args.excluded_dirs:
        for val in args.excluded_dirs:
            control_data['ignored_dirs'].append(val)
    logger.info('ignored_dirs: %s', ', '.join(control_data['ignored_dirs']))

    if args.altfile_digest:
        control_data['altfile_digest'] = args.altfile_digest.lower()
        if control_data['altfile_digest'] not in control_data['selected_digests']:
            logger.error('alt digest %s must be in selected digests',
                         control_data['altfile_digest'])
            return False
        control_data['altfile_name'] = '{}.{}.{}.{}'.format(
            output_title,
            control_data['altfile_digest'],
            output_tstamp,
            control_data['outfile_ext'],
        )
    return True


def main():
    """ Main entry point """
    control_data = dtconfig.CONTROL_DATA
    package_data = dtconfig.PACKAGE_DATA

    headline = f"{package_data['name']} Digester {package_data['version']}"

    print()
    print(headline)
    print()

    if not validate_args(headline):
        return False

    logger = logging.getLogger('_main_')

    header1 = [
        '#{}'.format('-' * 78),
        '#',
        '#  Base path: {}'.format(control_data['root_dir']),
        '#',
        '#{}'.format('-' * 78),
    ]
    header2 = [
        '#{}'.format('-' * 78),
        '',
    ]

    logger.debug('Logging out: %s', control_data['logfile_name'])
    logger.debug('Main output: %s', control_data['outfile_name'])

    outfile_header = '#         Digests               |'
    outfile_header += 'accessT |modifyT |createT |attr|watr|'
    outfile_header += '   size   |relative name'
    dtutils.outfile_write(
        control_data['outfile_name'],
        'w',
        header1 + [outfile_header] + header2,
    )
    altfile_header = '#        {} signature          |'
    altfile_header += 'accessT |modifyT |createT |watr|'
    altfile_header += '   size   |relative name'
    if control_data['altfile_digest']:
        logger.info('Alt  output: %s', control_data['altfile_name'])
        dtutils.outfile_write(
            control_data['altfile_name'],
            'w',
            header1 + [altfile_header.format(control_data['altfile_digest'])] + header2,
        )

    start_time = dtutils.curr_time_secs()
    logger.debug('MAINLINE starts - max_block_size=%d', control_data['max_block_size'])
    logger.debug('-;%s', dtdigester.fill_digest_str(control_data=control_data))
    walk_item = dtwalker.Walker()
    try:
        walk_item.initialize(control_data=control_data)
        start_walk_time = dtutils.curr_time_secs()
        walk_item.process_tree(control_data=control_data)
        end_walk_time = dtutils.curr_time_secs()
        walk_item.teardown(control_data=control_data)
    except KeyboardInterrupt:
        walk_item.teardown(control_data=control_data)
        logger.error('Ctrl+C pressed: exiting')
        logging.shutdown()
        return False
    end_time = dtutils.curr_time_secs()
    delta_time = end_time - start_time if end_time - start_time > 0 else 0.000001
    delta_walk_time = end_walk_time - start_walk_time if end_walk_time - start_walk_time > 0 else 0.000001
    logger.info(
        'run_time= %.3fs walk_time= %.3fs rate= %.2f MB/s bytes= %d',
        delta_time,
        delta_walk_time,
        control_data['counts']['bytes_read'] / 1024 / 1024 / delta_walk_time,
        control_data['counts']['bytes_read'],
    )
    footer = [
        '',
        '#{}'.format('-' * 78),
        '#',
        '#  Processed: {:,d} file(s), {:,d} folder(s) ({:,d} ignored, {:,d} errors) comprising {:,d} bytes'.format(
            control_data['counts']['files'],
            control_data['counts']['dirs'],
            control_data['counts']['ignored'],
            control_data['counts']['errors'],
            control_data['counts']['bytes_read'],
        ),
        '#',
        '#{}'.format('-' * 78),
    ]
    dtutils.outfile_write(control_data['outfile_name'], 'a', footer)
    if control_data['altfile_digest']:
        dtutils.outfile_write(control_data['altfile_name'], 'a', footer)
    logger.debug('MAINLINE ends - max_block_size=%d', control_data['max_block_size'])
    logger.info('Log ends')

    print()
    print(f"Logging out: {control_data['logfile_name']}")
    print(f"Main output: {control_data['outfile_name']}")



"""

    Copyright (c) 2017-2021 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

-----------------------------------------------------
    Don't print() directly from anything in here.
-----------------------------------------------------

"""

import logging  # For constants - do not log directly from here
import os
import queue

import dirtreedigest.utils as dtutils

if dtutils.shared_memory_available():
    from multiprocessing import shared_memory  # Python 3.8+
else:
    shared_memory = None


def reader_process(debug_queue, cmd_queue, results_queue, shm_mode, max_block_size):
    """ This is run as a subprocess, potentially with spawn()
        be careful with vars!
    """
    pid = os.getpid()
    buf_refs = {}
    buf_names = set()
    file_obj = None
    element = ''
    bytes_read = 0
    chunk = 0
    file_size = 0
    found_eof = False
    while True:
        try:
            try:
                cqi = cmd_queue.get_nowait()
                cmd = cqi.get('cmd', None)
            except queue.Empty:
                cmd = None
            if cmd == dtutils.Cmd.INIT:
                add_buf_names = cqi.get('buf_names', [])
                element = cqi.get('element', None)
                debug_queue.put((
                    logging.DEBUG,
                    f"READER: Init {add_buf_names} for element {element}"))
                buf_names.update(add_buf_names)
                for buf_name in buf_names:
                    if shm_mode and buf_name not in buf_refs:
                        buf_refs[buf_name] = shared_memory.SharedMemory(buf_name)
                bytes_read = 0
                found_eof = False
                chunk = 0
                if element:
                    errors = None
                    file_size = os.path.getsize(element)
                    try:
                        file_obj = open(element, 'rb')
                    except IOError as err:
                        errors = err
                        debug_queue.put((
                            logging.ERROR,
                            f"READER: Problem opening \"{element}\": {err}"))
                    results_queue.put({
                        'errors': errors,
                        'element': element,
                    })
            elif cmd == dtutils.Cmd.FREE:
                add_buf_names = cqi.get('buf_names', [])
                debug_queue.put((
                    logging.DEBUG,
                    f"READER: Free {add_buf_names}"))
                buf_names.update(add_buf_names)
            elif cmd == dtutils.Cmd.QUIT:
                # for buf in buf_refs:
                #     del(buf)
                if file_obj:
                    file_obj.close()
                debug_queue.put((
                    logging.INFO,
                    "READER: Quit"))
                break
            elif cmd:
                debug_queue.put((
                    logging.WARNING,
                    'reader_process() invalid command -- pid={} cmd={}'.format(pid, cmd)))
            else:  # Steady state
                if file_obj and not file_obj.closed:
                    block_size = min(max_block_size, file_size - bytes_read)
                    if (buf_names or not shm_mode) and file_size == 0:
                        found_eof = (bytes_read == file_size)
                        results_queue.put({
                            'chunk': chunk,
                            'block_size': block_size,
                            'buf_name': None,  # Shared memory mode
                            'buf_block': b'',  # Non-shared memory mode
                            'found_eof': found_eof,
                            'mbps': 0.0,
                            'element': element,
                        })
                    while (buf_names or not shm_mode) and block_size > 0:
                        start_time = dtutils.curr_time_secs()
                        buf_block = None
                        if shm_mode:
                            buf_name = buf_names.pop()
                            debug_queue.put((
                                logging.DEBUG,
                                f"READER: Reading chunk {chunk} of {block_size} bytes into {buf_name}"))
                            buf = buf_refs[buf_name]
                            buf.buf[:block_size] = file_obj.read(block_size)
                        else:
                            buf_name = None
                            debug_queue.put((
                                logging.DEBUG,
                                f"READER: Reading chunk {chunk} of {block_size} bytes into {buf_name}"))
                            buf_block = file_obj.read(block_size)
                        bytes_read += block_size
                        found_eof = (bytes_read == file_size)
                        end_time = dtutils.curr_time_secs()
                        delta_time = end_time - start_time if end_time - start_time > 0 else 0.000001
                        mbps = block_size / 1024 / 1024 / delta_time
                        results_queue.put({
                            'chunk': chunk,
                            'block_size': block_size,
                            'buf_name': buf_name,  # Shared memory mode
                            'buf_block': buf_block,  # Non-shared memory mode
                            'found_eof': found_eof,
                            'mbps': mbps,
                            'element': element,
                        })
                        chunk += 1
                        block_size = min(max_block_size, file_size - bytes_read)
                        if not shm_mode:
                            break
                    if found_eof:  # Makes sure the file always gets closed
                        file_obj.close()
        except KeyboardInterrupt:
            break



"""

    Copyright (c) 2017-2021 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import ctypes
import logging
import multiprocessing
import os
import stat
import sys

import dirtreedigest.digester as dtdigester
import dirtreedigest.reader as dtreader
import dirtreedigest.utils as dtutils
import dirtreedigest.worker as dtworker

if dtutils.shared_memory_available():
    from multiprocessing import shared_memory  # Python 3.8+
else:
    shared_memory = None


# pylint: disable=bad-whitespace
class Walker(object):
    """ Directory walker and supporting functions """

    # Windows file attributes
    FILE_ATTRIBUTE_NONE                = 0x00000  # noqa: E221
    FILE_ATTRIBUTE_READONLY            = 0x00001  # noqa: E221
    FILE_ATTRIBUTE_HIDDEN              = 0x00002  # noqa: E221
    FILE_ATTRIBUTE_SYSTEM              = 0x00004  # noqa: E221
    FILE_ATTRIBUTE_RESERVED_0008       = 0x00008  # noqa: E221
    FILE_ATTRIBUTE_DIRECTORY           = 0x00010  # noqa: E221
    FILE_ATTRIBUTE_ARCHIVE             = 0x00020  # noqa: E221
    FILE_ATTRIBUTE_DEVICE              = 0x00040  # noqa: E221
    FILE_ATTRIBUTE_NORMAL              = 0x00080  # noqa: E221
    FILE_ATTRIBUTE_TEMPORARY           = 0x00100  # noqa: E221
    FILE_ATTRIBUTE_SPARSE_FILE         = 0x00200  # noqa: E221
    FILE_ATTRIBUTE_REPARSE_POINT       = 0x00400  # noqa: E221
    FILE_ATTRIBUTE_COMPRESSED          = 0x00800  # noqa: E221
    FILE_ATTRIBUTE_OFFLINE             = 0x01000  # noqa: E221
    FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = 0x02000  # noqa: E221  # pylint: disable=invalid-name
    FILE_ATTRIBUTE_ENCRYPTED           = 0x04000  # noqa: E221
    FILE_ATTRIBUTE_INTEGRITY_STREAM    = 0x08000  # noqa: E221
    FILE_ATTRIBUTE_VIRTUAL             = 0x10000  # noqa: E221
    FILE_ATTRIBUTE_NO_SCRUB_DATA       = 0x20000  # noqa: E221
    FILE_ATTRIBUTE_INVALID             = 0x0FFFF  # noqa: E221
# pylint: enable=bad-whitespace

    def __init__(self):
        self.logger = logging.getLogger('walker')

    def _init_misc(self, control_data):
        """ Initialize items """
        control_data['debug_queue'] = multiprocessing.Queue()
        control_data['ignored_file_pats'] = dtutils.compile_patterns(
            control_data['ignored_files'],
            control_data['ignore_path_case'],
        )
        control_data['ignored_dir_pats'] = dtutils.compile_patterns(
            control_data['ignored_dirs'],
            control_data['ignore_path_case'],
        )

    def _start_shared_memory(self, control_data):
        """ Initialize shared memory """
        control_data['buffer_blocks'] = []
        control_data['buffer_sizes'] = []
        control_data['buffer_names'] = []
        for _ in range(control_data['max_buffers']):
            if control_data['shm_mode']:
                buf = shared_memory.SharedMemory(create=True, size=control_data['max_block_size'])
                buffer_name = buf.name
                control_data['buffer_blocks'].append(buf)
                control_data['buffer_names'].append(buffer_name)
            else:
                control_data['buffer_blocks'].append([None])
            control_data['buffer_sizes'].append(0)

    def _end_shared_memory(self, control_data):
        """ Clean up shared memory """
        if control_data['shm_mode']:
            for i in range(control_data['max_buffers']):
                shm_buf = shared_memory.SharedMemory(name=control_data['buffer_names'][i])
                shm_buf.close()
                shm_buf.unlink()

    def _start_reader(self, control_data):
        """ Start long-running worker processes
            Until subprocessed are ended, raising exceptions can hang the parent process
        """
        control_data['reader_proc'] = None
        control_data['reader_cmd_queue'] = multiprocessing.Queue()
        control_data['reader_results_queue'] = multiprocessing.Queue()
        reader_proc = multiprocessing.Process(
            target=dtreader.reader_process,
            args=(
                control_data['debug_queue'],
                control_data['reader_cmd_queue'],
                control_data['reader_results_queue'],
                control_data['shm_mode'],
                control_data['max_block_size'],
            ),
        )
        reader_proc.name = '---Reader'
        reader_proc.start()

    def _end_reader(self, control_data):
        """ End reader subprocess """
        control_data['reader_cmd_queue'].put({
            'cmd': dtutils.Cmd.QUIT,
        })
        import time
        time.sleep(0.5)  # Because reasons
        while not control_data['reader_results_queue'].empty():
            retval = control_data['reader_results_queue'].get()
            self.logger.debug('Draining queue: %s', retval)
        dtutils.flush_debug_queue(control_data['debug_queue'], logging.getLogger('reader'))
        while control_data['reader_proc']:
            if control_data['reader_proc'] and not control_data['reader_proc'].is_alive():
                control_data['reader_proc'] = None

    def _start_workers(self, control_data):
        """ Start long-running worker processes
            Until subprocessed are ended, raising exceptions can hang the parent process
        """
        control_data['worker_procs'] = []
        control_data['worker_cmd_queues'] = []
        control_data['worker_results_queue'] = multiprocessing.Queue()
        for i in range(len(control_data['selected_digests'])):
            control_data['worker_cmd_queues'].append(multiprocessing.JoinableQueue())
            worker_proc = multiprocessing.Process(
                target=dtworker.worker_process,
                args=(
                    control_data['debug_queue'],
                    control_data['worker_cmd_queues'][i],
                    control_data['worker_results_queue'],
                    control_data['shm_mode'],
                ),
            )
            worker_proc.name = f'---Worker-{i}'
            worker_proc.start()
            control_data['worker_procs'].append(worker_proc)

    def _end_workers(self, control_data):
        """ End worker subprocesses """
        for worker_cmd_queue in control_data['worker_cmd_queues']:
            worker_cmd_queue.put({
                'cmd': dtutils.Cmd.QUIT,
            })
        while not control_data['worker_results_queue'].empty():
            retval = control_data['worker_results_queue'].get()
            self.logger.debug('Draining queue: %s', retval)
        dtutils.flush_debug_queue(control_data['debug_queue'], logging.getLogger('worker'))
        while any(control_data['worker_procs']):
            for i in range(len(control_data['selected_digests'])):
                if (control_data['worker_procs'][i] is not None) and (
                        not control_data['worker_procs'][i].is_alive()):
                    self.logger.debug(
                        'join %d (state = %s) at %f',
                        i,
                        control_data['worker_procs'][i].is_alive(),
                        dtutils.curr_time_secs(),
                    )
                    control_data['worker_procs'][i].join()
                    control_data['worker_procs'][i] = None

    def initialize(self, control_data):
        self._init_misc(control_data)
        self._start_shared_memory(control_data)
        self._start_reader(control_data)
        self._start_workers(control_data)

    def teardown(self, control_data):
        self._end_shared_memory(control_data)
        self._end_reader(control_data)
        self._end_workers(control_data)

    def get_win_filemode(self, elem):
        """ Windows: get system-specific file stats """
        return ctypes.windll.kernel32.GetFileAttributesW(elem)

    def is_win_symlink(self, elem):
        """ Windows: system-specific symlink check """
        return os.path.isdir(elem) and (
            self.get_win_filemode(elem) & self.FILE_ATTRIBUTE_REPARSE_POINT)

    def process_tree(self, control_data):
        """ Process the given directory tree """
        results = []
        self._walk_tree(
            control_data=control_data,
            root_dir=control_data['root_dir'],
            callback=self.visit_element,
            results=results)
        return results

    def _walk_tree(self, control_data, root_dir, callback, results):
        """ Re-entrant directory tree walker """
        try:
            dir_list = os.listdir(root_dir)
        except FileNotFoundError:
            self.logger.warning('FileNotFoundError %s', root_dir)
            control_data['counts']['errors'] += 1
            return
        except NotADirectoryError:
            self.logger.warning('NotADirectoryError %s', root_dir)
            control_data['counts']['errors'] += 1
            return
        except PermissionError:
            self.logger.warning('PermissionError %s', root_dir)
            control_data['counts']['errors'] += 1
            return
        print("Processing path {}".format(root_dir))
        for elem in sorted(dir_list):
            pathname = dtutils.unixify_path(os.path.join(root_dir, elem))
            # print("Processing file {}".format(pathname))
            try:
                stats = os.lstat(pathname)
            except FileNotFoundError:
                self.logger.warning('FileNotFoundError %s', root_dir)
                control_data['counts']['errors'] += 1
                continue
            if stat.S_ISDIR(stats.st_mode):
                if dtutils.elem_is_matched(
                        root_dir,
                        pathname,
                        control_data['ignored_dir_pats']):
                    self.logger.info('D IGNORED %s', pathname)
                    control_data['counts']['ignored'] += 1
                    continue
                results.append(self.visit_element(control_data, pathname, stats))
                self._walk_tree(
                    control_data=control_data,
                    root_dir=pathname,
                    callback=callback,
                    results=results)
            elif stat.S_ISREG(stats.st_mode):
                if dtutils.elem_is_matched(
                        root_dir,
                        pathname,
                        control_data['ignored_file_pats']):
                    self.logger.info('F IGNORED %s', pathname)
                    control_data['counts']['ignored'] += 1
                    continue
                results.append(self.visit_element(control_data, pathname, stats))
            else:
                results.append(self.visit_element(control_data, pathname, stats))
            dtutils.flush_debug_queue(control_data['debug_queue'], logging.getLogger('worker'))

    def visit_element(self, control_data, element, stats):
        """ Stat / digest a specific element found during the directory walk """
        root_dir = control_data['root_dir']
        elem_data = {}
        elem_data['name'] = dtutils.get_relative_path(root_dir, dtutils.unixify_path(element))
        elem_data['mode'] = stats.st_mode
        if sys.platform == 'win32':
            elem_data['mode_w'] = self.get_win_filemode(element)
        else:
            elem_data['mode_w'] = self.FILE_ATTRIBUTE_NONE
        elem_data['size'] = stats[stat.ST_SIZE]
        elem_data['atime'] = stats[stat.ST_ATIME]
        elem_data['mtime'] = stats[stat.ST_MTIME]
        elem_data['ctime'] = stats[stat.ST_CTIME]
        elem_data['digests'] = None
        elem_data['type'] = '?'
        alt_digest_len = 1
        if control_data['altfile_digest']:
            alt_digest_len = dtdigester.DIGEST_FUNCTIONS[control_data['altfile_digest']]['len']
        if sys.platform == 'win32' and self.is_win_symlink(element):
            elem_data['type'] = 'J'
            alt_digest = '?' * alt_digest_len
            sorted_digests = dtdigester.fill_digest_str(control_data, 'x')
        elif stat.S_ISDIR(stats.st_mode):
            elem_data['type'] = 'D'
            elem_data['size'] = 0
            alt_digest = '-' * alt_digest_len
            sorted_digests = dtdigester.fill_digest_str(control_data, '-')
            control_data['counts']['dirs'] += 1
        elif stat.S_ISREG(stats.st_mode):
            elem_data['type'] = 'F'
            elem_data['digests'] = dtdigester.digest_file(control_data, element)
            if elem_data['digests']:
                control_data['counts']['files'] += 1
                sorted_digests = '{' + ', '.join('{}: {}'.format(
                    i, elem_data['digests'][i]) for i in sorted(
                        elem_data['digests'])) + '}'
                if control_data['altfile_digest']:
                    alt_digest = elem_data['digests'][control_data['altfile_digest']]
            else:
                self.logger.warning('F Problems processing %s', element)
                control_data['counts']['errors'] += 1
                sorted_digests = dtdigester.fill_digest_str(control_data, '!')
                alt_digest = '!' * alt_digest_len
        else:
            elem_data['type'] = '?'
            sorted_digests = dtdigester.fill_digest_str(control_data, '?')
        file_details = '{};{};{:08x};{:08x};{:08x};{:04x};{:04x};{:010x};{}'.format(
            elem_data['type'],
            sorted_digests,
            elem_data['atime'], elem_data['mtime'], elem_data['ctime'],
            elem_data['mode'], elem_data['mode_w'],
            elem_data['size'],
            elem_data['name'])
        alt_digest = '-' * alt_digest_len
        if elem_data['type'] == 'D':
            alt_digest = '-' * alt_digest_len
        elif not elem_data['digests']:
            alt_digest = '?' * alt_digest_len
        elif control_data['altfile_digest']:
            alt_digest = elem_data['digests'][control_data['altfile_digest']]
        alt_details = '{};{:08x};{:08x};{:08x};{:04x};{:010x};{}'.format(
            alt_digest,
            elem_data['atime'], elem_data['mtime'], elem_data['ctime'],
            elem_data['mode_w'],
            elem_data['size'],
            elem_data['name'])
        self.logger.debug('%s', file_details)
        dtutils.outfile_write(control_data['outfile_name'], 'a', [
            '{}'.format(file_details),
        ])
        if control_data['altfile_digest']:
            dtutils.outfile_write(control_data['altfile_name'], 'a', [
                '{}'.format(alt_details),
            ])
        return elem_data



"""

    Copyright (c) 2017-2021 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import hashlib
import logging
import queue
import zlib

import dirtreedigest.utils as dtutils

if dtutils.shared_memory_available():
    from multiprocessing import shared_memory  # Python 3.8+
else:
    shared_memory = None


class CsumNoop(object):
    """ No-op digester """
    name = 'noop'

    def __init__(self):
        self.checksum = 0

    def update(self, msg):
        """ Update checksum """
        pass

    def hexdigest(self):
        """ Return digest string in hexadecimal format """
        return '{0:0{1}x}'.format(self.checksum, 8)


class CsumNoop0(CsumNoop):
    """ No-op digester """
    name = 'noop0'


class CsumNoop1(CsumNoop):
    """ No-op digester """
    name = 'noop1'


class CsumNoop2(CsumNoop):
    """ No-op digester """
    name = 'noop2'


class CsumNoop3(CsumNoop):
    """ No-op digester """
    name = 'noop3'


class CsumNoop4(CsumNoop):
    """ No-op digester """
    name = 'noop4'


class CsumNoop5(CsumNoop):
    """ No-op digester """
    name = 'noop5'


class CsumNoop6(CsumNoop):
    """ No-op digester """
    name = 'noop6'


class CsumNoop7(CsumNoop):
    """ No-op digester """
    name = 'noop7'


class CsumAdler32(object):
    """ Adler32 digester """
    name = 'adler32'

    def __init__(self):
        self.checksum = 1

    def update(self, msg):
        """ Update checksum """
        self.checksum = zlib.adler32(msg, self.checksum)

    def hexdigest(self):
        """ Return digest string in hexadecimal format """
        return '{0:0{1}x}'.format(self.checksum, 8)


class CsumCrc32(object):
    """ CRC32 digester """
    name = 'crc32'

    def __init__(self):
        self.checksum = 0

    def update(self, msg):
        """ Update checksum """
        self.checksum = zlib.crc32(msg, self.checksum)

    def hexdigest(self):
        """ Return digest string in hexadecimal format """
        return '{0:0{1}x}'.format(self.checksum, 8)


# pylint: disable=bad-whitespace, no-member
DIGEST_FUNCTIONS_TEST = {
    # No-op test functions
    'noop':       {'name': 'noop',      'len':   8, 'entry': CsumNoop},  # noqa: E241
    'noop0':      {'name': 'noop0',     'len':   8, 'entry': CsumNoop0},  # noqa: E241
    'noop1':      {'name': 'noop1',     'len':   8, 'entry': CsumNoop1},  # noqa: E241
    'noop2':      {'name': 'noop2',     'len':   8, 'entry': CsumNoop2},  # noqa: E241
    'noop3':      {'name': 'noop3',     'len':   8, 'entry': CsumNoop3},  # noqa: E241
    'noop4':      {'name': 'noop4',     'len':   8, 'entry': CsumNoop4},  # noqa: E241
    'noop5':      {'name': 'noop5',     'len':   8, 'entry': CsumNoop5},  # noqa: E241
    'noop6':      {'name': 'noop6',     'len':   8, 'entry': CsumNoop6},  # noqa: E241
    'noop7':      {'name': 'noop7',     'len':   8, 'entry': CsumNoop7},  # noqa: E241
}
DIGEST_FUNCTIONS_MAIN = {
    # Digesters in Python since at least Python 2.7
    'crc32':      {'name': 'crc32',     'len':   8, 'entry': CsumCrc32},  # noqa: E241
    'adler32':    {'name': 'adler32',   'len':   8, 'entry': CsumAdler32},  # noqa: E241
    'md5':        {'name': 'md5',       'len':  32, 'entry': hashlib.md5},  # noqa: E241
    'sha1':       {'name': 'sha1',      'len':  40, 'entry': hashlib.sha1},  # noqa: E241
    'sha224':     {'name': 'sha224',    'len':  56, 'entry': hashlib.sha224},  # noqa: E241
    'sha256':     {'name': 'sha256',    'len':  64, 'entry': hashlib.sha256},  # noqa: E241
    'sha384':     {'name': 'sha384',    'len':  96, 'entry': hashlib.sha384},  # noqa: E241
    'sha512':     {'name': 'sha512',    'len': 128, 'entry': hashlib.sha512},  # noqa: E241
}
DIGEST_FUNCTIONS_PY36 = {
    # Digesters added in Python 3.6
    # Variable length digests are excluded (e.g., 'shake_128')
    'blake2b':    {'name': 'blake2b',   'len': 128, 'entry': hashlib.blake2b},  # noqa: E241
    'blake2s':    {'name': 'blake2s',   'len':  64, 'entry': hashlib.blake2s},  # noqa: E241
    'sha3_224':   {'name': 'sha3_224',  'len':  56, 'entry': hashlib.sha3_224},  # noqa: E241
    'sha3_256':   {'name': 'sha3_256',  'len':  64, 'entry': hashlib.sha3_256},  # noqa: E241
    'sha3_384':   {'name': 'sha3_384',  'len':  96, 'entry': hashlib.sha3_384},  # noqa: E241
    'sha3_512':   {'name': 'sha3_512',  'len': 128, 'entry': hashlib.sha3_512},  # noqa: E241
}
# pylint: enable=bad-whitespace, no-member
DIGEST_FUNCTIONS = {}
DIGEST_FUNCTIONS.update(DIGEST_FUNCTIONS_TEST)
DIGEST_FUNCTIONS.update(DIGEST_FUNCTIONS_MAIN)
DIGEST_FUNCTIONS.update(DIGEST_FUNCTIONS_PY36)


''' From least to most secure, excluding tests '''
DIGEST_PRIORITY = [
    'crc32',
    'adler32',
    'md5',
    'sha1',
    'sha224',
    'sha3_224',
    'sha256',
    'blake2s',
    'sha3_256',
    'sha384',
    'sha3_384',
    'sha512',
    'blake2b',
    'sha3_512',
]


def validate_digests(control_data):
    """ returns list of available digests """
    logger = logging.getLogger('digester')
    if not control_data['selected_digests']:
        return None
    digests_available = set(DIGEST_FUNCTIONS.keys()) & set(control_data['selected_digests'])
    digests_not_found = set(control_data['selected_digests']) - digests_available
    if digests_not_found:
        logger.warning('Warning: invalid digest(s): %s', digests_not_found)
    digest_list = sorted(digests_available)
    if len(digest_list) > control_data['max_concurrent_jobs']:
        logger.error(
            'Error: Number of digests (%d) may not exceed max jobs (%d)',
            len(digest_list), control_data['max_concurrent_jobs'])
        return None
    logger.debug('Digests: %s', digest_list)
    return digest_list


def fill_digest_str(control_data, fillchar='-'):
    """ Create a padded dummy digest value """
    return '{' + ', '.join('{}: {}'.format(
        i, fillchar * DIGEST_FUNCTIONS[i]['len']) for i in sorted(
            control_data['selected_digests'])) + '}'


def digest_file(control_data, element):
    """ Digest a given element """
    logger = logging.getLogger('digester')
    start_time = dtutils.curr_time_secs()
    logger.debug('process_file(%s)', element)
    total_jobs = len(control_data['selected_digests'])
    bytes_read = 0
    found_eof = False
    hash_stats = {}

    control_data['reader_cmd_queue'].put({
        'cmd': dtutils.Cmd.INIT,
        'buf_names': control_data['buffer_names'],
        'element': element,
    })
    result = control_data['reader_results_queue'].get()
    dtutils.flush_debug_queue(control_data['debug_queue'], logging.getLogger('worker'))
    if result['errors']:
        return hash_stats

    for i, worker_cmd_queue in enumerate(control_data['worker_cmd_queues']):
        digest_name = control_data['selected_digests'][i]
        digest_func = DIGEST_FUNCTIONS[digest_name]['entry']
        worker_cmd_queue.put({
            'cmd': dtutils.Cmd.INIT,
            'digest_func': digest_func,
            'element': element,
        })

    while not found_eof:
        block_read = control_data['reader_results_queue'].get()
        logger.debug('BLOCK READ: %s %s %s', block_read['block_size'], block_read['buf_name'], element)
        found_eof = block_read['found_eof']
        block_size = block_read['block_size']
        buf_block = block_read['buf_block']
        buf_name = block_read['buf_name']
        bytes_read += block_size
        for worker_cmd_queue in control_data['worker_cmd_queues']:
            worker_cmd_queue.put({
                'cmd': dtutils.Cmd.PROCESS,
                'block_size': block_size,
                'buf_name': buf_name,  # Shared memory mode
                'buf_block': buf_block,  # Non-shared memory mode
                'element': element,
            })
        # Gather any intermediate outputs
        jobs = total_jobs
        while jobs > 0:
            try:
                result = control_data['worker_results_queue'].get_nowait()
            except queue.Empty:
                continue
            if result:
                jobs -= 1
        if buf_name:
            control_data['reader_cmd_queue'].put({
                'cmd': dtutils.Cmd.FREE,
                'buf_names': [buf_name],
            })

    for worker_cmd_queue in control_data['worker_cmd_queues']:
        worker_cmd_queue.put({
            'cmd': dtutils.Cmd.RESULT,
        })
    dtutils.flush_debug_queue(control_data['debug_queue'], logging.getLogger('worker'))
    jobs = total_jobs
    while jobs > 0:
        retval = control_data['worker_results_queue'].get()
        logger.debug('RETVAL: %s', retval)
        if 'msg' in retval:
            continue  #TODO: messages pop up here from time to time.
        hash_stats[retval['digest_name']] = retval['digest_value']
        jobs -= 1
    dtutils.flush_debug_queue(control_data['debug_queue'], logging.getLogger('worker'))

    end_time = dtutils.curr_time_secs()
    delta_time = end_time - start_time if end_time - start_time > 0 else 0.000001
    logger.debug('MAINLINE finished at %f', end_time)
    logger.debug(
        'run_time= %.3fs rate= %.3f MB/s bytes= %d %s',
        delta_time,
        bytes_read / 1024 / 1024 / delta_time,
        bytes_read,
        element)
    control_data['counts']['bytes_read'] += bytes_read
    return hash_stats



"""

    Copyright (c) 2017-2021 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import logging
import re

from collections import defaultdict
from datetime import datetime
from enum import Enum
from os.path import basename, dirname

import dirtreedigest.digester as dtdigester


class DiffType(Enum):
    M_UNDEF = 0  # Undefined (error case)  # noqa: E221
    M_NONE  = 1  # Diff file Name, Diff Data ("no match" anywhere)  # noqa: E221
    M_SFSD  = 2  # Same Full name, Same Data (Opposite only)  # noqa: E221
    M_SFDD  = 3  # Same Full name, Diff Data (Opposite only)  # noqa: E221
    M_SNSD  = 4  # Same file Name, Same Data (Opposite and Same sides)  # noqa: E221
    M_DNSD  = 5  # Diff file Name, Same Data (Opposite and Same sides)  # noqa: E221


class Comparator(object):
    """ Digest blob comparator and supporting functions """
    elements_l = []
    elements_r = []
    files_by_name_l = {}
    files_by_name_r = {}
    basepath_l = ''
    basepath_r = ''
    best_digest = None
    control_data = None

    def __init__(self, control_data):
        self.logger = logging.getLogger('comparator')
        self.control_data = control_data

    def read_dtd_report(self, filename):
        element_pat = re.compile(
            r"^(.+?);{(.+?)};(.+?);(.+?);(.+?);(.+?);(.+?);(.+?);(.*)$")
        legacy_pat = re.compile(
            r"^(.+?);(.+?);(.+?);(.+?);(.+?);(.+?);(.*)$")
        basepath_pat = re.compile(
            r"^#\s+Base path:\s+(.*)$")
        elements = []
        basepath = ''
        with open(filename, 'r', encoding='utf-8') as fileh:
            for line in fileh:
                line = line.rstrip('\n')
                elem = {}
                mval = element_pat.match(line)
                if mval:
                    elem['digests'] = {}
                    for digestpair in mval[2].split(','):
                        (digest, val) = digestpair.strip().split(':')
                        elem['digests'][digest.strip()] = val.strip()
                    elem['type'] = mval[1]
                    elem['atime'] = mval[3]
                    elem['mtime'] = mval[4]
                    elem['ctime'] = mval[5]
                    elem['attr_std'] = mval[6]
                    elem['attr_win'] = mval[7]
                    elem['size'] = mval[8]
                    elem['full_name'] = mval[9]
                    elem['dir_name'] = dirname(elem['full_name'])
                    elem['file_name'] = basename(elem['full_name'])
                else:  # Legacy
                    mval = legacy_pat.match(line)
                    if mval:
                        elem['digests'] = {}
                        elem['digests']['md5'] = mval[1]
                        elem['type'] = 'F'
                        if elem['digests']['md5'].startswith('?'):
                            elem['type'] = '?'
                        elif elem['digests']['md5'].startswith('-'):
                            elem['type'] = 'D'
                        elem['atime'] = mval[2]
                        elem['mtime'] = mval[3]
                        elem['ctime'] = mval[4]
                        elem['attr_std'] = '0000'
                        elem['attr_win'] = mval[5]
                        elem['size'] = mval[6]
                        elem['full_name'] = mval[7]
                        elem['dir_name'] = dirname(elem['full_name'])
                        elem['file_name'] = basename(elem['full_name'])
                    else:
                        mval = basepath_pat.match(line)
                        if mval:
                            basepath = mval[1]
                if elem:
                    if elem['type'] not in ['D', 'F']:
                        self.logger.warning(f"Ignoring file type '{elem['type']}' for {elem['full_name']}")
                    else:
                        elem['id'] = len(elements)
                        elements.append(elem)
        return (basepath, elements)

    def choose_best_digest_for_compare(self, elems1, elems2):
        best = -1
        if not len(elems1) > 0 and len(elems2) > 0:
            self.logger.error('Cannot choose a digest - not found')
            return None
        if 'digests' not in elems1[0] and 'digests' in elems2[0]:
            self.logger.error('Cannot choose a digest - not found')
            return None
        s1 = set(elems1[0]['digests'])
        s2 = set(elems2[0]['digests'])
        for digest in (s1 & s2):
            if digest in dtdigester.DIGEST_PRIORITY:
                best = max(best, dtdigester.DIGEST_PRIORITY.index(digest))
        if best < 0:
            self.logger.error('Cannot choose a digest - none in common')
            return None
        best_name = dtdigester.DIGEST_PRIORITY[best]
        return best_name

    def slice_data(self, elements):
        files_by_name = {elem['full_name']: elem for elem in elements if elem['type'] == 'F'}
        files_by_digest = defaultdict(list)
        for elem in elements:
            if elem['type'] == 'F':
                cmp_digest = elem['digests'][self.best_digest]  #TODO: Mix in file size here?
                files_by_digest[cmp_digest].append(elem)
        return (files_by_name, files_by_digest)

    def compare_by_full_names(self, name_same):
        elems_changed = []
        for name in name_same:
            digest_l = self.files_by_name_l[name]['digests'][self.best_digest]
            digest_r = self.files_by_name_r[name]['digests'][self.best_digest]
            self.files_by_name_l[name]['match'] = [self.files_by_digest_r[digest_l]]
            self.files_by_name_r[name]['match'] = [self.files_by_digest_l[digest_r]]
            if digest_l == digest_r:
                self.files_by_name_l[name]['status'] = 'same'
                self.files_by_name_r[name]['status'] = 'same'
                if self.files_by_name_l[name]['mtime'] != self.files_by_name_r[name]['mtime']:
                    time_l = datetime.fromtimestamp(int("0x"+self.files_by_name_l[name]['mtime'], 16))
                    time_r = datetime.fromtimestamp(int("0x"+self.files_by_name_r[name]['mtime'], 16))
                    self.logger.info("SAME-T: %ss: %s", int((time_r - time_l).total_seconds()), name)
            else:
                self.files_by_name_l[name]['status'] = 'changed'
                self.files_by_name_r[name]['status'] = 'changed'
                if self.files_by_name_l[name]['mtime'] == self.files_by_name_r[name]['mtime']:
                    self.logger.info("MOD-T : %s", name)
                elems_changed.append(self.files_by_name_r[name])
        return (elems_changed)

    def check_lhs(self, name_diff_l):
        elems_moved = []
        elems_deleted = []
        for name in name_diff_l:
            digest_l = self.files_by_name_l[name]['digests'][self.best_digest]
            # print("checking", name)
            if digest_l in self.files_by_digest_r:
                # print(name, digest_l)
                # matched_elems = [elem['full_name'] for elem in self.files_by_digest_r[digest_l]]
                # matched_names = ','.join(matched_elems)
                # print("< MOVED   {} == {}".format(name, matched_names))
                self.files_by_name_l[name]['status'] = 'moved'
                self.files_by_name_l[name]['match'] = self.files_by_digest_r[digest_l]
                elems_moved.append(self.files_by_name_l[name])
            else:
                # print("< DELETED {}".format(name))
                self.files_by_name_l[name]['status'] = 'deleted'
                elems_deleted.append(self.files_by_name_l[name])
        return (elems_moved, elems_deleted)

    def check_rhs(self, name_diff_r):
        elems_copied = []
        elems_added = []
        for name in name_diff_r:
            digest_r = self.files_by_name_r[name]['digests'][self.best_digest]
            # print("checking", name)
            if digest_r in self.files_by_digest_l:
                # print(name, digest_r)
                # matched_elems = [elem['full_name'] for elem in self.files_by_digest_l[digest_r]]
                # matched_names = ','.join(matched_elems)
                # print("> COPIED  {} == {}".format(name, matched_names))
                self.files_by_name_r[name]['status'] = 'copied'
                for elem in self.files_by_digest_l[digest_r]:
                    if elem['file_name'] == self.files_by_name_r[name]['file_name']:
                        self.logger.info("------: Found likely source: %s", self.files_by_name_r[name]['full_name'])
                        break
                self.files_by_name_r[name]['match'] = self.files_by_digest_l[digest_r]
                elems_copied.append(self.files_by_name_r[name])
            else:
                # print("> ADDED   {}".format(name))
                self.files_by_name_r[name]['status'] = 'added'
                elems_added.append(self.files_by_name_r[name])
        return (elems_copied, elems_added)

    def compare(self, file_l, file_r):
        """ Main entry: compare two dirtreedigest reports """
        (self.basepath_l, self.elements_l) = self.read_dtd_report(file_l)
        (self.basepath_r, self.elements_r) = self.read_dtd_report(file_r)

        self.logger.info("Root L: %s", self.basepath_l)
        self.logger.info("Root R: %s", self.basepath_r)

        self.best_digest = self.choose_best_digest_for_compare(self.elements_l, self.elements_r)
        if self.best_digest is None:
            return None

        self.logger.info("BestDG: %s", self.best_digest)

        (self.files_by_name_l, self.files_by_digest_l) = self.slice_data(self.elements_l)
        (self.files_by_name_r, self.files_by_digest_r) = self.slice_data(self.elements_r)

        name_set_l = set(self.files_by_name_l)
        name_set_r = set(self.files_by_name_r)
        name_diff_l = name_set_l - name_set_r
        name_diff_r = name_set_r - name_set_l
        name_same = name_set_r & name_set_l

        (elems_changed) = self.compare_by_full_names(name_same)
        (elems_moved, elems_deleted) = self.check_lhs(name_diff_l)
        (elems_copied, elems_added) = self.check_rhs(name_diff_r)

        for elem in sorted(elems_changed, key=lambda k: k['full_name']):
            self.logger.info(f"MOD   : \"{elem['full_name']}\"")

        for elem in sorted(elems_added, key=lambda k: k['full_name']):
            self.logger.info(f"ADD   : \"{elem['full_name']}\"")

        for elem in sorted(elems_deleted, key=lambda k: k['full_name']):
            self.logger.info(f"DEL   : \"{elem['full_name']}\"")

        for elem in sorted(elems_copied, key=lambda k: k['full_name']):
            self.logger.info(f"COPY  : \"{elem['full_name']}\" == \"{'---'}\"")

        for elem in sorted(elems_moved, key=lambda k: k['full_name']):
            self.logger.info(f"MOVE  : \"{elem['full_name']}\" == \"{'---'}\"")

        self.logger.info("ElemsL: %d", len(self.elements_l))
        self.logger.info("ElemsR: %d", len(self.elements_r))
        self.logger.info("FilesL: %d", len(self.files_by_name_l))
        self.logger.info("FilesR: %d", len(self.files_by_name_r))
        self.logger.info("  Both: %d", len(name_same))
        self.logger.info("Only L: %d", len(name_diff_l))
        self.logger.info("Only R: %d", len(name_diff_r))



"""

    Copyright (c) 2017-2021 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

-----------------------------------------------------
    Don't print() directly from anything in here.
-----------------------------------------------------

"""

import logging  # For constants - do not log directly from here
import os
import queue

import dirtreedigest.utils as dtutils

if dtutils.shared_memory_available():
    from multiprocessing import shared_memory  # Python 3.8+
else:
    shared_memory = None


def worker_process(debug_queue, cmd_queue, results_queue, shm_mode):
    """ This is run as a subprocess, potentially with spawn()
        be careful with vars!
    """
    pid = os.getpid()
    digest_name = 'None'
    buf_refs = {}
    while True:
        try:
            try:
                cqi = cmd_queue.get()
                cmd = cqi.get('cmd', None)
            except queue.Empty:
                cmd = None
            if cmd == dtutils.Cmd.INIT:
                digest_func = cqi.get('digest_func', None)
                if digest_func:
                    digest_instance = digest_func()
                    digest_name = digest_instance.name
                else:
                    digest_name = 'None'
                    debug_queue.put((
                        logging.ERROR,
                        'worker_process({}) init -- pid={} Missing constructor'.format(
                            digest_name, pid)))
                debug_queue.put((
                    logging.DEBUG,
                    'worker_process({}) init -- pid={}'.format(
                        digest_name, pid)))
                cmd_queue.task_done()
            elif cmd == dtutils.Cmd.PROCESS:
                block_size = cqi.get('block_size', None)
                buf_name = cqi.get('buf_name', None)
                if shm_mode:
                    if block_size > 0:
                        if buf_name not in buf_refs:
                            buf_refs[buf_name] = shared_memory.SharedMemory(buf_name)
                        buf = buf_refs[buf_name]
                        byte_block = buf.buf[:block_size]
                        debug_queue.put((
                            logging.DEBUG,
                            'worker_process() reading shared memory -- pid={} l={} c={} d={}'.format(
                                pid, block_size, byte_block[0], digest_name)))
                    else:
                        byte_block = b''
                else:
                    byte_block = cqi.get('buf_block', None)
                digest_instance.update(byte_block)
                if shm_mode:
                    del(byte_block)  # Otherwise shared_memory spews `BufferError: cannot close exported pointers exist`
                debug_queue.put((
                    logging.DEBUG,
                    'worker_process() process -- pid={} buf_name={} l={} d={}'.format(
                        pid, buf_name, block_size, digest_instance.hexdigest())))
                results_queue.put({
                    'msg': '{} processed'.format(digest_name),
                })
                cmd_queue.task_done()
            elif cmd == dtutils.Cmd.RESULT:
                debug_queue.put((
                    logging.DEBUG,
                    'worker_process({}) result -- pid={} digest={}'.format(
                        digest_name, pid, digest_instance.hexdigest())))
                results_queue.put({
                    'digest_name': digest_name,
                    'digest_value': digest_instance.hexdigest(),
                })
                cmd_queue.task_done()
            elif cmd == dtutils.Cmd.QUIT:
                # for buf in buf_refs:
                #     del(buf)
                debug_queue.put((
                    logging.DEBUG,
                    'worker_process({}) quit -- pid={}'.format(
                        digest_name, pid)))
                cmd_queue.task_done()
                break
            else:
                debug_queue.put((
                    logging.WARNING,
                    'worker_process() invalid command -- pid={} cmd={}'.format(pid, cmd)))
        except KeyboardInterrupt:
            break



"""

    Copyright (c) 2017-2021 Martin F. Falatic

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import argparse
import logging
import sys

from pathlib import Path

import dirtreedigest.__config__ as dtconfig
import dirtreedigest.comparator as dtcompare
import dirtreedigest.utils as dtutils


def validate_args(headline):
    """ Validate command-line arguments """
    control_data = dtconfig.CONTROL_DATA
    # package_data = dtconfig.PACKAGE_DATA

    epilog = ''

    parser = argparse.ArgumentParser(
        # description=package_data['description'],
        epilog=epilog)
    parser.add_argument('file_left', nargs='?',
                        default=None, type=str, action='store',
                        help='First digest file')
    parser.add_argument('file_right', nargs='?',
                        default=None, type=str, action='store',
                        help='Second digest file')
    parser.add_argument('--title', dest='output_title', metavar='TITLE',
                        default=None, type=str, action='store',
                        help='alternate output title')
    parser.add_argument('--notimestamps', dest='notimestamps',
                        action='store_true',
                        help='ignore timestamps')
    parser.add_argument('--noattributes', dest='noattributes',
                        action='store_true',
                        help='ignore attributes')
    parser.add_argument('--nocase', dest='nocase',
                        action='store_true',
                        help='case insensitive matching')
    parser.add_argument('--debug', dest='debug',
                        action='store_true',
                        help='more debugging to the logfile')
    args = parser.parse_args()

    if not (args.file_left and args.file_right):
        parser.print_help()
        return False

    control_data['file_l'] = args.file_left
    control_data['file_r'] = args.file_right

    control_data['logfile_level'] = logging.INFO
    if args.debug:
        control_data['logfile_level'] = logging.DEBUG

    control_data['outfile_suffix'] = 'cmp'

    if args.output_title:
        output_title = args.output_title
    else:
        output_title = '{}-{}.{}'.format(
            control_data['outfile_prefix'],
            control_data['outfile_suffix'],
            f"{Path(control_data['file_l']).stem} vs {Path(control_data['file_r']).stem}",
        )

    control_data['outfile_name'] = '{}.{}'.format(
        output_title,
        'txt',
    )

    control_data['logfile_name'] = '{}.{}'.format(
        output_title,
        control_data['logfile_ext'],
    )

    dtutils.start_logging(
        control_data['logfile_name'],
        control_data['logfile_level'],
        control_data['console_level'],
    )

    logger = logging.getLogger('_main_')
    logger.info('Log begins')
    logger.info('-' * 78)
    logger.info(headline)
    logger.info(f"Using Python {sys.version}")
    logger.info('-' * 78)

    logger.info('Digest file L: %s', args.file_left)
    logger.info('Digest file R: %s', args.file_right)

    control_data['notimestamps'] = False
    if args.notimestamps:
        control_data['notimestamps'] = True
    logger.info('notimestamps: %s', control_data['notimestamps'])

    control_data['noattributes'] = False
    if args.notimestamps:
        control_data['noattributes'] = True
    logger.info('noattributes: %s', control_data['noattributes'])

    control_data['ignore_path_case'] = False
    if args.nocase:
        control_data['ignore_path_case'] = True
    logger.info('ignore_path_case: %s', control_data['ignore_path_case'])

    return True


def main():
    """ Main entry point """
    control_data = dtconfig.CONTROL_DATA
    package_data = dtconfig.PACKAGE_DATA

    headline = f"{package_data['name']} Comparator {package_data['version']}"

    print()
    print(headline)
    print()

    if not validate_args(headline=headline):
        return False

    logger = logging.getLogger('_main_')

    header1 = [
        '#{}'.format('-' * 78),
        '#',
        '#  Path L: {}'.format('x'),
        '#  Path R: {}'.format('y'),
        '#',
        '#{}'.format('-' * 78),
    ]
    header2 = [
        '#{}'.format('-' * 78),
        '',
    ]
    logger.info('Main output: %s', control_data['outfile_name'])
    outfile_header = '#         Digests               |'
    outfile_header += 'accessT |modifyT |createT |attr|watr|'
    outfile_header += '   size   |relative name'
    dtutils.outfile_write(
        control_data['outfile_name'],
        'w',
        header1 + [outfile_header] + header2,
    )
    # start_time = dtutils.curr_time_secs()
    logger.debug('MAINLINE starts')

    comparator = dtcompare.Comparator(control_data=control_data)

    comparator.compare(
        file_l=control_data['file_l'],
        file_r=control_data['file_r'],
    )

    # end_time = dtutils.curr_time_secs()
    # delta_time = end_time - start_time
    footer = [
        '',
        '#{}'.format('-' * 78),
        '#',
        '#  Processed: {:,d} file(s), {:,d} folder(s) ({:,d} ignored, {:,d} errors) comprising {:,d} bytes'.format(
            control_data['counts']['files'],
            control_data['counts']['dirs'],
            control_data['counts']['ignored'],
            control_data['counts']['errors'],
            control_data['counts']['bytes_read'],
        ),
        '#',
        '#{}'.format('-' * 78),
    ]
    dtutils.outfile_write(control_data['outfile_name'], 'a', footer)
    logger.debug('MAINLINE ends')
    logger.info('Log ends')

    print()
    print(f"Main output: {control_data['outfile_name']}")
